#include <stdio.h>
#include <stdlib.h>
/*#include <math.h>*/
#include <mpi.h>

/*
 * This example presents the implementation of the graph partitioning
 * known as the combinatorial optimization problem 'Max-Cut', which is
 * 'NP-complete'. The target is to partition the set of vertices of
 * an edge-weigted undirected graph into two disjoint sets S1 and S2
 * so, that the sum of the weights of all edges connecting S1 and S2
 * is maximal.
 * The edge-weigted undirected graph is represented by an symetric
 * adjacency matrix which represents an input data.
 * The result of the partitioning is formed as an 0/1 value vector,
 * where value 0 denotes a vertex belonging to ithe set S1 and
 * value 1 denotes a vertex belonging to the set S2.
 */
/*
 * Algorithm:   analytical - generates recursively all possible cut
 *              vectors and calculates its cut. The vector having the
 *              maximal cut value is chosen as the solution.
 *              There can exist more than one solution with the same
 *              cut value.
 * Input:       adjmatrix17.dat
 * Output:      pmaxcut.out
 */

int calculate_cut(MPI_Comm comm, int msize, int *adjmatrix, FILE *fpout);
void convert2bin(unsigned long n, int pos, int *cutset);

int main(int argc, char **argv)
{
    MPI_Comm comm;
    FILE *fpin, *fpout = NULL;
    int  msize, *adjmatrix;
    int  i, j, k;
    int  nprocs, me, root = 0;
    int  fin, error = 0;
    char *fnamin;
    char *fnamout = "pmaxcut.out";

    MPI_Init(&argc, &argv);
    MPI_Comm_size(MPI_COMM_WORLD, &nprocs);
    MPI_Comm_rank(MPI_COMM_WORLD, &me);
    comm = MPI_COMM_WORLD;

    /* adjmatrix17.dat: msize=17  nn=65535 */
    /* adjmatrix32.dat: msize= 32  nn= 2147483647 */

    if (me == root) {
        if (argc < 2) {
            fprintf(stderr, "Error: Data matrix missing.\n");
            error++;
        }
        else {
            fnamin = argv[1];
            fpin = fopen(fnamin, "r");
            if (fpin == NULL) {
                fprintf(stderr, "Error: %s file can not be opened.\n", fnamin);

                error++;
            }
        }
        fpout = fopen(fnamout, "w");
        if (fpout == NULL) {
            fprintf(stderr, "Error: %s file can not be opened.\n", fnamout);
            error++;
        }
    }
    MPI_Bcast(&error, 1, MPI_INT, root, comm);
    if (error) {
        MPI_Finalize();
        return 1;
    }

    if (me == root) {
        fscanf(fpin, "%d", &msize);
        fprintf(stdout, "Calculation MAXCUT begin ...\n");
        fprintf(stdout, "  Data matrix: %s\n", fnamin);
        fprintf(stdout, "  msize= %d  nprocs= %d\n", msize, nprocs);
    }
    MPI_Bcast(&msize, 1, MPI_INT, root, comm);

    adjmatrix = (int *) malloc(sizeof(int) * msize*msize);
    if (adjmatrix == NULL) {
        error++;
    }
    (void) MPI_Allreduce(&error, &k, 1, MPI_INT, MPI_SUM, comm);
    error = k;
    if (error) {
        if (me == root) {
            fprintf(stderr, "Error: Memory allocation (main).\n");
            fclose(fpin);
            fclose(fpout);
        }
        MPI_Finalize();
        return 1;
    }

    /* Read input data */
    if (me == root) {
        for (i=0; i<msize; i++) {
            k = i*msize;
            for (j=0; j<msize; j++) {
                fscanf(fpin, "%d", &adjmatrix[k+j]);
            }
        }
    }
    MPI_Bcast(adjmatrix, msize*msize, MPI_INT, root, comm);

    error = calculate_cut(comm, msize, adjmatrix, fpout);

    if (me == root) {
        fclose(fpin);
        fin = fclose(fpout);
        if (fin != 0) {
            fprintf(stderr, "Error: %s file can not be closed.\n", fnamout);
            error++;
        }
        else {
            fprintf(stdout, "Calculation MAXCUT finished.\n");
        }
        fflush(stdout);
        fflush(stderr);
    }
    free(adjmatrix);
    MPI_Bcast(&error, 1, MPI_INT, root, comm);
    MPI_Finalize();
    return error;
}

int calculate_cut(MPI_Comm comm, int msize, int *adjmatrix, FILE *fpout)
{
    int nprocs, me, root = 0;
    int i, j, k, m, pos;
    int cut, maxcut;
    int *cutset, *mcuta, *work;
    unsigned long mn, *mna, *lwork;
    unsigned long n, nn, lb, ub;
    int error = 0;

    MPI_Comm_size(comm, &nprocs);
    MPI_Comm_rank(comm, &me);

    cutset = (int *) malloc(sizeof(int) * msize);
    mna   = (unsigned long *) malloc(sizeof(unsigned long) * nprocs * 2);
    lwork = (unsigned long *) malloc(sizeof(unsigned long) * nprocs * 2);
    mcuta = (int *) malloc(sizeof(int) * nprocs);
    work  = (int *) malloc(sizeof(int) * nprocs);
    if ((cutset == NULL)||
        (mna == NULL)||(lwork == NULL)||
        (mcuta == NULL)||(work == NULL)) {
        error++;
    }
    (void) MPI_Allreduce(&error, &k, 1, MPI_INT, MPI_SUM, comm);
    error = k;
    if (error) {
        if (me == root) {
            fprintf(stderr, "Error: Memory allocation (calculate_cut).\n");
        }
        return 1;
    }

    /* nn = 2^(msize-1) - 1 */
    nn = 1;
    for (i = 1; i < msize; i++) {
        nn = nn * 2;
    }
    nn = nn -1;

    /* work distribution */
    /* n = ceiling(dble(nn) / dble(nprocs)) */
    n = (nn / nprocs) + ((nn % nprocs) ? 1 : 0);
    lb = me * n;
    ub = (me+1)*n-1;
    if (nn < ub)
        ub = nn;

    /* Debugging ... */
   for (i=0; i<nprocs*2; i++) {
        mna[i] = 0;
        lwork[i] = 0;
    }
    mna[me] = lb;
    mna[me+nprocs] = ub;
    (void) MPI_Allreduce(mna, lwork, nprocs*2, MPI_UNSIGNED_LONG, MPI_SUM, comm
);
    if (me == root) {
        fprintf(stdout, "  Calculating local cut ...(nn= %lu)\n", nn);
        fprintf(stdout, "  Work distribution:\n");
        fprintf(stdout, "    block size= %lu\n", n);
        for (i=0; i<nprocs; i++) {
            fprintf(stdout, "    [P%d]  lb= %lu  ub= %lu\n",
                    i, lwork[i], lwork[i+nprocs]);
        }
    }
    /* ... Debugging */

    for (i=0; i<nprocs*2; i++) {
        mna[i] = 0;
    }
    for (i=0; i<nprocs; i++) {
        mcuta[i] = 0;
    }
    pos = msize-1;

    for (n=lb; n<=ub; n++) {
        for (i=0; i<msize; i++)
            cutset[i] = 0;
        convert2bin(n, pos, cutset);
        cut = 0;
        for (i=0; i<msize; i++) {
            k = i*msize;
            for (j=i+1; j<msize; j++) {
                m = k+j;
                if ((adjmatrix[m] != 0) &&
                    (cutset[i] != cutset[j])) {
                    cut += adjmatrix[m];
                }
            }
        }
        if (mcuta[me] < cut) {
            mcuta[me] = cut;
            mna[me] = n;
        }
    }
    if (me == root) {
        fprintf(stdout, "  Max-Reduction on local cuts\n");
    }
    (void) MPI_Allreduce(mcuta, work, nprocs, MPI_INT, MPI_SUM, comm);
    (void) MPI_Allreduce(mna, lwork, nprocs, MPI_UNSIGNED_LONG, MPI_SUM, comm);

    for (i=0; i<nprocs; i++) {
        mcuta[i] = work[i];
        mna[i] = lwork[i];
    }
    mn = mna[0];
    maxcut = mcuta[0];
    for (i=0; i<nprocs; i++) {
        if (maxcut < mcuta[i]) {
            maxcut = mcuta[i];
            mn = mna[i];
        }
    }
    for (i=0; i<msize; i++)
       cutset[i] = 0;
    convert2bin(mn, pos, cutset);

    if (me == root) {
        fprintf(fpout, "Calculation MAXCUT: msize= %d  nn= %lu\n", msize, nn);
        fprintf(fpout, "MaxCut Value= %d\n", maxcut);
        fprintf(fpout, "MaxCut Set:  ");
        for (i=0; i<msize-1; i++)
            fprintf(fpout, "%d, ", cutset[i]);
        fprintf(fpout, "%d\n", cutset[msize-1]);
        fflush(fpout);
    }

    MPI_Barrier(comm);
    free(mcuta);
    free(mna);
    free(work);
    free(lwork);
    free(cutset);

    return 0;
}

void convert2bin(unsigned long n, int pos, int *cutset)
{
    if (n == 0)
        return;
    if (n == 1) {
        cutset[pos] = 1;
        return;
    }
    cutset[pos] = n % 2;
    convert2bin(n/2, pos-1, cutset);
}

