#include <stdlib.h>
#include <stdio.h>

/*
 * This example presents the implementation of the graph partitioning
 * known as the combinatorial optimization problem 'Max-Cut', which is
 * 'NP-complete'. The target is to partition the set of vertices of
 * an edge-weigted undirected graph into two disjoint sets S1 and S2
 * so, that the sum of the weights of all edges connecting S1 and S2
 * is maximal.
 * The edge-weigted undirected graph is represented by an symetric
 * adjacency matrix which represents an input data.
 * The result of the partitioning is formed as an 0/1 value vector,
 * where value 0 denotes a vertex belonging to ithe set S1 and
 * value 1 denotes a vertex belonging to the set S2.
 */
/*
 * Algorithm:   analytical - generates recursively all possible cut
 *              vectors and calculates its cut. The vector having the
 *              maximal cut value is chosen as the solution.
 *              There can exist more than one solution with the same
 *              cut value.
 * Input:       adjmatrix17.dat
 * Output:      maxcut.out
 */

int  calculate_cut(int msize, int *adjmatrix, FILE *fpout);
void convert2bin(unsigned long n, int pos, int *cutset);

int main(int argc, char **argv)
{
    FILE *fpin, *fpout;
    int msize, *adjmatrix;
    int i, j, k;
    int fin;
    int error = 0;
    char *fnamin;
    char *fnamout = "maxcut.out";

    /* adjmatrix17.dat:  msize= 17  nn= 65535 */
    /* adjmatrix24.dat": msize= 24  nn= 8388607 */
    /* adjmatrix32.dat:  msize= 32  nn= 2147483647 */
    /* file name of adjmatrix */

    if (argc < 2) {
        fprintf(stderr, "Error: Input data matrix missing.\n");
        return(1);
    }
    fnamin = argv[1];
    fprintf(stdout, "Input data matrix: %s\n", fnamin);
    if (argc > 2) {
        fnamout = argv[2];
    }
    fprintf(stdout, "Output data: %s\n", fnamout);


    fpin = fopen(fnamin, "r");
    if (fpin == NULL) {
        fprintf(stderr, "Error: %s file can not be opened.\n", fnamin);
        return(1);
    }
    fscanf(fpin, "%d", &msize);
    adjmatrix = (int *) malloc(sizeof(int) * msize*msize);
    if (adjmatrix == NULL) {
        fprintf(stderr, "Error: Memory allocation.\n");
        fclose(fpin);
        return(1);
    }

    for (i = 0; i < msize; i++) {
        k = i*msize;
        for (j = 0; j < msize; j++) {
            fscanf(fpin, "%d", &adjmatrix[k+j]);
        }
    }
    fclose(fpin);

    fpout = fopen(fnamout, "w");
    if (fpout == NULL) {
        fprintf(stderr, "Error: %s file can not be opened.\n", fnamout);
        return(1);
    }
    error = calculate_cut(msize, adjmatrix, fpout);
    free(adjmatrix);

    fin = fclose(fpout);
    if (fin != 0) {
        fprintf(stderr, "Error: %s file can not be closed.\n", fnamout);
        error++;
    }

    fprintf(stdout, "Calculation MAXCUT finished\n");
    return(error);
}

int calculate_cut(int msize, int *adjmatrix, FILE *fpout)
{
    int i, j, k, m, pos;
    double dn;
    unsigned long nn, nm, n;
    int maxcut, cut;
    int *cutset;

/*
    dn = pow(2, msize-1) - 1;
    //nn = static_cast<unsigned long long>(dn);
    nn = (unsigned long long)dn;
*/
    nn = 1;
    for (i = 1; i < msize; i++) {
        nn = nn * 2;
    }
    nn = nn -1;
    fprintf(stdout, "Calculation MAXCUT: msize= %d  nn= %lu\n", msize, nn);

    cutset = (int *) malloc(sizeof(int) * msize);
    if (cutset == NULL) {
        fprintf(stderr, "Error: Memory allocation\n");
        return(1);
    }

    maxcut = 0;
    nm = 0;
    pos = msize-1;
    cut = 0;
    for (n = 0; n <= nn; n++) {
        for (i = 0; i < msize; i++)
            cutset[i] = 0;
        convert2bin(n, pos, cutset);
        cut = 0;
        for (i = 0; i < msize; i++) {
            k = i*msize;
            for (j = i+1; j < msize; j++) {
                m = k+j;
                if ((adjmatrix[m] != 0)&&
                    (cutset[i] != cutset[j])) {
                    cut += adjmatrix[m];
                }
            }
        }
        if (maxcut < cut) {
            maxcut = cut;
            nm = n;
        }
    }

    for (i = 0; i < msize; i++)
        cutset[i] = 0;
    convert2bin(nm, pos, cutset);

    fprintf(fpout, "Calculation MAXCUT: msize= %d  nn= %lu\n", msize, nn);
    fprintf(fpout, "maxcut= %ld\n", maxcut);
    fprintf(fpout, "maxcutset:  ");
    for (i=0; i<msize-1; i++)
        fprintf(fpout, "%d, ", cutset[i]);
    fprintf(fpout, "%d\n", cutset[msize-1]);

    free(cutset);
    return(0);
}
 
void convert2bin(unsigned long n, int pos, int *cutset)
{
    if (n == 0)
        return;

    if (n == 1) {
        cutset[pos] = 1;
        return;
    }
    cutset[pos] = n % 2;
    convert2bin(n/2, pos-1, cutset);
}

